/*
** Module   : ScreenBuffer.java
** Abstract : screen buffer to transferring state to/from the thin client
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 EVL 20050827   @22377 Creation based on general concept. Necessary
**                           modification to get compiled.
** 002 SIY 20050831   @22544 Minor formatting fixes.
** 003 EVL 20050908   @22615 Adding key code to be transferable via
**                           network.
** 004 NVS 20051118   @23402 Added new method getWidgetValue(int widgetID)
** 005 NVS 20060307   @24939 Added state flags and methods to work with
**                           state changes.
** 006 NVS 20060309   @24968 Added methods to produce changed only entries
**                           and fixed setState() to remove the old value
**                           when the new state is UNINITIALIZED.
** 007 NVS 20060314   @25070 Added getBaseId() and reset() methods.
** 008 SIY 20060316   @25138 Added support for frame headers.
** 009 GES 20060627   @27603 Added support for dynamic down frame size. 
** 010 GES 20060628   @27693 Clear all changed flags at end of getChanged
**                           to ensure that subsequent processing doesn't
**                           detect more values as changed when they have
**                           already been sent to the other side.
** 011 GES 20060707   @27907 Added support for dynamic titles.
** 012 SIY 20060720   @28144 Added support for frame value and active 
**                           widget ID.
** 013 GES 20060801   @28287 Converted to Externalizable to improve
**                           efficiency because Serializable has unneeded
**                           processing (for versioning...) that is
**                           executed on "inflation".
** 014 GES 20060802   @28413 Moved some logic to an external class to
**                           promote reuse.
** 015 SIY 20060810   @28544 Fixed initialization of maps.
** 016 GES 20061101   @30893 Added clearHeaders().
** 017 SIY 20061117   @31289 Updated for support of dynamic row/column. 
** 018 SIY 20070418   @33098 Resetting ENTERED state now is limited to
**                           specified list of IDs.
** 019 NVS 20070927   @35259 Added debugging aid method view(). Useful in
**                           the debugging session to print the contents
**                           of the screen buffer with "invoke view()".
** 020 NVS 20071018   @35482 Enhanced debug output for BaseDataType case.
** 021 SIY 20071127   @36035 Refreshed to use type safe collections.
**                           Minor cleanup in javadoc.
** 022 GES 20071211   @36330 Added frame line support.
** 023 SIY 20080320   @37601 Added support for screen values for widgets
**                           which can have non-matching data and screen
**                           values. 
** 024 SIY 20080320   @37698 Added resetting of the UNINITIALIZED when 
**                           merging changes.
** 025 GES 20090422   @41902 Converted to standard string formatting.
** 026 SIY 20111101          Updated imports, moved from CHUI package.
** 027 EVL 20140822          Rework for widget ID usage approach.
** 028 EVL 20140926          Adding methods to handle incoming new dynamic widgets.
** 029 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 030 GES 20150115          Added isUninitialized().
*/
 
package com.goldencode.p2j.ui;

import java.io.*;
import java.util.*;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.util.BaseDataType;

/**
 * This is the base object to interchange server and thin client with the
 * current widget values.
 * <p>
 * The reviewed implementation adds more control over the contents of the
 * buffer. The total number of widgets is stored as well as the frame ID. This
 * information allows for calculating widget index into a byte array that 
 * keeps widget state: uninitialized or changed.
 * <p>
 * Only those widgets marked as changed are present in the buffer to minimize
 * data exchange.
 */
public class ScreenBuffer
implements Externalizable
{
   /** State flag */
   public static final byte STATE_MIN = 0;

   /** State flag */
   public static final byte UNCHANGED = 0;

   /** State flag */
   public static final byte UNINITIALIZED = 1;

   /** State flag */
   public static final byte CHANGED = 2;

   /** State flag */
   public static final byte ENTERED = 4;

   /** State flag */
   public static final byte STATE_MAX = UNINITIALIZED + CHANGED + ENTERED;

   /** Frame ID where this screen buffer belongs */
   private int baseId = -1;

   /** Maximum number of entries */
   private int size = -1;

   /**
    * Per widget state map.  This map must not be accessed directly to put, remove or get values.
    * Use helper methods for respective operation.
    */
   private Map<Integer, Byte> state = new HashMap<Integer, Byte>();

   /** The map to store (widget ID, current value) pairs */
   private Map<Integer, Object> buffer = new HashMap<Integer, Object>();

   /** The map to store header widget values. */
   private Map<Integer, Object> headers = new HashMap<Integer, Object>();

   /** The map for some screen values. */
   private Map<Integer, Object> values = new HashMap<Integer, Object>();
   
   /** Key code which will be transferred in key related events. */
   private int keyCode = -1;

   /** Rows in a dynamic down frame or -1 if no runtime override exists. */
   private int down = -1;

   /** Current frame line or -1 if it is undefined. */
   private int line = -1;

   /** Value of dynamically calculated frame column. */
   private int column = BaseConfig.INV_COORD;
   
   /** Value of dynamically calculated frame row. */
   private int row = BaseConfig.INV_COORD;
   
   /** Dynamic title text or <code>null</code> for no dynamic support. */
   private String title = null;

   /** Value of the last active widget. */
   private String frameValue = null;
   
   /** ID of the last active widget. */
   private int widgetId = -1;
   
   /**
    * The default constructor creates a rudimentary ScreenBuffer which remains
    * empty.
    */
   public ScreenBuffer()
   {
      size = 0;
   }
   
   /**
    * Constructor.
    * Prepares internal structures to take the data.
    *
    * @param    frameId
    *           the ID of the frame where this ScreenBuffer belongs
    * @param    numWidgets
    *           the total number of widgets in the frame
    */
   public ScreenBuffer(int frameId, int numWidgets)
   {
      size = numWidgets;
      baseId = frameId;
   }

   /**
    * Constructor.
    * Prepares internal structures to take the data.
    *
    * @param    frameId
    *           the ID of the frame where this ScreenBuffer belongs
    * @param    ids
    *           the array of the widget IDs to put inside the buffer
    */
   public ScreenBuffer(int frameId, int[] ids)
   {
      baseId = frameId;
      if (ids != null)
      {
         size = ids.length;
         for (int i = 0; i < ids.length; i++)
         {
            setStateInt(ids[i], UNINITIALIZED);
         }
      }
      else
      {
         size = 0;
      }
   }

   /**
    * Gets the maximum size of this screen buffer.
    *
    * @return   The total number of widgets in the frame
    */
   public int size()
   {
      return size;
   }

   /**
    * Gets the frame ID from this screen buffer.
    *
    * @return   The frame ID.
    */
   public int getBaseId()
   {
      return baseId;
   }

   /**
    * Adding new widget into screen buffer for the given widget ID.
    * 
    * @param    widgetID
    *           The ID of the widget to be added.
    */
   public void addWidget(int widgetID)
   {
       // store initilal state
      setStateInt(widgetID, UNINITIALIZED);
      // and increase the size
      size++;
   }
   
   /**
    * Removing existed widget from screen buffer for the given widget ID.
    * 
    * @param    widgetID
    *           The ID of the widget to be added.
    */
   public void removeWidget(int widgetID)
   {
      // verify the widget ID is valid for this buffer
      if (!isIdValid(widgetID))
         return;

      // remove the state entry
      state.remove(widgetID);
      // remove the value itself
      buffer.remove(widgetID);
      // and decrease the total size
      size--;
   }
   
   /**
    * Storing the new value for the given widget ID.
    * 
    * @param    widgetID
    *           The ID of the widget to which the data is related.
    * @param    value
    *           New value as an Object instance.
    */
   public void putWidgetValue(int widgetID, Object value)
   {
      // verify the widget ID is valid for this buffer
      if (!isIdValid(widgetID))
         return;

      // if widget already changed state override initial
      byte stateNew = getStateInt(widgetID);
      
      // change state
      stateNew &= STATE_MAX - UNINITIALIZED - ENTERED;
      stateNew |= CHANGED;
      // store new state
      setStateInt(widgetID, stateNew);
      // and store new widget value
      buffer.put(widgetID, value);
   }

   /**
    * Storing the new header value for the given widget ID.
    * 
    * @param    widgetID
    *           The ID of the widget to which the data is related.
    * @param    value
    *           New value as an Object instance.
    */
   public void putHeaderValue(int widgetID, Object value)
   {
      if (headers == null)
         headers = new LinkedHashMap<Integer, Object>();
      
      headers.put(widgetID, value);
   }
   
   /**
    * Gets the current state for the given widget ID.
    * 
    * @param    widgetID
    *           The ID of the widget to which the data is related.
    * @return   State of this widget or <code>-1</code> if such ID does not
    *           exist.
    */
   public byte getState(int widgetID)
   {
      // verify the widget ID is valid for this buffer
      if (!isIdValid(widgetID))
      {
         return -1;
      }
      
      // get the current status
      return getStateInt(widgetID); 
   }

   /**
    * Stores the new state for the widget with the given widget ID.
    * IF the new state is UNINITIALIZED, the previous value of the widget,
    * if any, is deleted.
    * 
    * @param    widgetID
    *           The ID of the widget to which the data is related.
    * @param    state
    *           new state
    */
   public void setState(int widgetID, byte state)
   {
      // verify the widget ID is valid for this buffer
      if (!isIdValid(widgetID))
      {
         return;
      }

      if (state < STATE_MIN || state > STATE_MAX)
         return;

      // remove widget value if become uninitialized
      if (state == UNINITIALIZED)
      {
         buffer.remove(new Integer(widgetID));
      }
      
      // store new state value for this widget ID
      setStateInt(widgetID, (byte)state);
   }
   
   /**
    * Requests the iterator to walk through the content of the screen buffer.
    * 
    * @return   The iterator for the screen buffer content.
    */
   public Iterator<Integer> getWidgetIDIterator()
   {
      return buffer.keySet().iterator();
   }

   /**
    * Requests the iterator to walk through the content of the header widgets.
    * 
    * @return   The iterator for the screen buffer content.
    */
   public Iterator<Integer> getHeaderIDIterator()
   {
      if (headers == null)
         return (new HashMap<Integer, Object>()).keySet().iterator();
      
      return headers.keySet().iterator();
   }
   
   /**
    * Getting the value in the Object form for the given widget ID.
    * 
    * @param    widgetID
    *           The numeric ID of a widget whose value must be returned.
    * 
    * @return   The Object instance of the widget value.
    */
   public Object getWidgetValue(int widgetID)
   {
      return buffer.get(widgetID);
   }

   /**
    * Getting the header value in the Object form for the given widget ID.
    * 
    * @param    widgetID
    *           The numeric ID of a widget whose value must be returned.
    * 
    * @return   The Object instance of the widget value.
    */
   public Object getHeaderValue(int widgetID)
   {
      if (headers == null)
         return null;
      
      return headers.get(widgetID);
   }
   
   /**
    * Set screen value for specified widget.
    * 
    * @param   id
    *          Widget id.
    * @param   screenValue
    *          Widget screen value.
    */
   public void putScreenValue(int id, Object screenValue)
   {
      values.put(id, screenValue);
   }
   
   /**
    * Get current screen value.
    * 
    * @param   id
    *          Widget id.
    *          
    * @return  Current widget value.
    */
   public Object getScreenValue(int id)
   {
      return values.get(id);
   }
   
   /**
    * Stores the key code for a client-server transfer.
    * 
    * @param    keyCode
    *           The P2J key code for the current event.
    */
   public void putKeyCode(int keyCode)
   {
      this.keyCode = keyCode;
   }
   
   /**
    * Extracts the key code for the current event.
    * 
    * @return   The P2J key code value for the current event.
    */
   public int getKeyCode()
   {
      return keyCode;
   }

   /**
    * Stores the caller-specified number of rows for the associated down
    * frame for client-server transfer.  If a non-positive number is 
    * specified, it will be converted to 0 (which means that the maximum 
    * possible value should be calculated by the client and used).
    * 
    * @param    down
    *           The number of rows in the frame or 0 for forcing a calculated
    *           value or -1 for no dynamic override of the down size.
    */
   public void putDown(int down)
   {
      this.down = down;
   }
   
   /**
    * Extracts the caller-specified number of rows for the associated down
    * frame.
    * 
    * @return   The number of rows in the frame or 0 for forcing a calculated
    *           value or -1 for no dynamic override of the down size.
    */
   public int getDown()
   {
      return down;
   }
   
   /**
    * Stores the current <code>FRAME-LINE</code> for the given frame.
    * 
    * @param    line
    *           The frame line or -1 for undefined.
    */
   public void putLine(int line)
   {
      this.line = line;
   }
   
   /**
    * Extracts the current <code>FRAME-LINE</code> for the given frame. 
    * 
    * @return   The frame line or -1 for undefined.
    */
   public int getLine()
   {
      return line;
   }
   
   /**
    * Store new value for dynamically calculated frame row.
    * 
    * @param    row
    *           The new value for the frame row or -1 for no explicit value.
    */
   public void putRow(int row)
   {
      this.row = row;
   }
   
   /**
    * Get current value for dynamically calculated frame row.
    * 
    * @return   The current value for the row.
    */
   public int getRow()
   {
      return row;
   }
   
   /**
    * Store new value for dynamically calculated frame column.
    * 
    * @param    column
    *           The new value for the frame column or -1 for no explicitly
    *           set value.
    */
   public void putColumn(int column)
   {
      this.column = column;
   }
   
   /**
    * Get current value for dynamically calculated frame column.
    * 
    * @return   The current value for the column.
    */
   public int getColumn()
   {
      return column;
   }
   
   /**
    * Sets the caller-specified dynamic title for the associated frame.
    * 
    * @param    title
    *           The current value of the dynamic title text or
    *           <code>null</code> if no dynamic title support exists for this
    *           frame.
    */
   public void putTitle(String title)
   {
      this.title = title;
   }
   
   /**
    * Extracts the caller-specified dynamic title for the associated frame.
    * 
    * @return   The current value of the dynamic title text or
    *           <code>null</code> if no dynamic title support exists for this
    *           frame.
    */
   public String getTitle()
   {
      return title;
   }

   /**
    * Gets a new <code>ScreenBuffer</code> where only widgets with the
    * changed values are present.  As a side effect the current instance's
    * <code>CHANGED</code> flags are reset.
    * 
    * @return   new differential <code>ScreenBuffer</code>
    */
   public ScreenBuffer getChanged()
   {
      ScreenBuffer diff = new ScreenBuffer(baseId, size);
      
      // copy down value
      diff.down = this.down;
      
      // copy frame line value
      diff.line = this.line;
      
      // copy row and column
      diff.row    = this.row;
      diff.column = this.column;
      
      // copy title value
      diff.title = this.title;
      
      // copying state info
      copyState(diff);

      // copy changed widget values
      Iterator<Integer> iter = buffer.keySet().iterator();
      
      Integer id = null;
      Object value = null;
      
      while (iter.hasNext())
      {
         id = iter.next();

         // get current state and check for CHANGED flag
         if (!((getStateInt(id) & CHANGED) == CHANGED))
         {
            continue;
         }
         
         value = buffer.get(id);
         diff.buffer.put(id, value);
      }

      // copy header widgets 
      if (headers != null)
      {
         iter = headers.keySet().iterator();
         
         while(iter.hasNext())
         {
            id = iter.next();
            diff.putHeaderValue(id, headers.get(id));
         }
      }
      
      // reset changed flags in source (otherwise recursive processing may
      // be incorrectly obtaining the same values as changed when they have
      // already been sent)
      resetChanged();
      
      return diff;
   }

   /**
    * Merges this ScreenBuffer with the changed widgets from the specified
    * one. Changed widgets are marked as ENTERED.
    * 
    * @param    diff
    *           a differential <code>ScreenBuffer</code> to copy widget values
    *           from.
    */
   public void mergeChanged(ScreenBuffer diff)
   {
      // copy changed widget values
      Iterator<Integer> iter = diff.buffer.keySet().iterator();
      
      Integer id = null;
      Object value = null;
      
      while (iter.hasNext())
      {
         id = iter.next();

         // get current status from diff buffer and check for CHANGED flag
         if (!((diff.getStateInt(id) & CHANGED) == CHANGED))
            continue;
         
         value = diff.buffer.get(id);
         this.buffer.put(id, value);
         // get current state from this current buffer
         byte stateCurr = getStateInt(id);
         // reset UNINITIALIZED
         stateCurr &= ~UNINITIALIZED;
         stateCurr |= ENTERED;
         setStateInt(id, stateCurr);
      }
      
      // copy the frame line
      this.line = diff.line;
      
      // copy the last entered key
      this.keyCode = diff.keyCode;
      
      values.clear();
      values.putAll(diff.values);
   }
 
   /**
    * Resets all widgets having ENTERED state.
    * @param    ids 
    *           Array of IDs of entered widgets.
    */
   public void resetEntered(int[] ids)
   {
      byte flags = STATE_MAX - ENTERED;

      if (ids == null)
      {
         Iterator<Integer> iter = state.keySet().iterator();

         while (iter.hasNext())
         {
            Integer id = iter.next();
            // reset entered state bit for next widget
            bitwiseAndState(id.intValue(), flags);
         }
      }
      else
      {
         for (int i = 0; i < ids.length; i++)
         {
            if (isIdValid(ids[i]))
            {
               // reset entered state bit for next widget
               bitwiseAndState(ids[i], flags);
            }
         }
      }
   }

   /**
    * Resets all widgets having CHANGED state.
    */
   public void resetChanged()
   {
      byte flags = STATE_MAX - CHANGED;
      Iterator<Integer> iter = state.keySet().iterator();

      while (iter.hasNext())
      {
         Integer id = iter.next();
         // reset changed state bit for next widget
         bitwiseAndState(id.intValue(), flags);
      }
   }

   /**
    * Resets all widgets to the UNINITIALIZED state.
    */
   public void reset()
   {
      Iterator<Integer> iter = state.keySet().iterator();

      while (iter.hasNext())
      {
         setStateInt(iter.next(), UNINITIALIZED);
      }
      
      buffer.clear();
   }

   /**
    * Checks the ENTERED state of a widget.
    *
    * @param    widgetID
    *           The ID of the widget to which the data is related.
    * @return   <code>true</code> if the widget has been modified during the 
    *           last input operation.
    */
   public boolean isEntered(int widgetID)
   {
      // verify the widget ID is valid for this buffer
      if (!isIdValid(widgetID))
      {
         return false;
      }

      // verify the state has ENTERED flag on
      return (getStateInt(widgetID) & ENTERED) == ENTERED;
   }

   /**
    * Checks the UNINITIALIZED state of a widget.
    *
    * @param    widgetID
    *           The ID of the widget to which the data is related.
    *
    * @return   <code>true</code> if the widget is in the uninitialized state.
    */
   public boolean isUninitialized(int widgetID)
   {
      // verify the widget ID is valid for this buffer
      if (!isIdValid(widgetID))
      {
         // false means it is initialized which is worse, but this doesn't make much sense
         // either
         return true;
      }

      // verify the state has UNINITIALIZED flag on
      return (getStateInt(widgetID) & UNINITIALIZED) == UNINITIALIZED;
   }

   /**
    * Get the value of the last active widget in the frame.
    * 
    * @return   Returns the frameValue.
    */
   public String getFrameValue()
   {
      return frameValue;
   }

   /**
    * Save the value of the last active widget in the frame.
    *   
    * @param    frameValue 
    *           The frame value to set.
    */
   public void setFrameValue(String frameValue)
   {
      this.frameValue = frameValue;
   }

   /**
    * Get ID of the last active widget.
    * 
    * @return   The widget ID or <code>-1</code> if no such widget was
    *           detected.
    */
   public int getValueWidgetId()
   {
      return widgetId;
   }

   /**
    * Save the ID of last active widget.
    *          
    * @param    widgetID
    *           The widgetId to set.
    */
   public void setValueWidgetId(int widgetID)
   {
      widgetId = widgetID;
   }
   
   /**
    * Removes all header values from the buffer.
    */
   public void clearHeaders()
   {
      if (headers != null)
      {
         headers = null;
      }
   }
   
   /**
    * Debugging Aid.
    * Visualizes the contents of this instance.
    *
    * @return   The array of strings that represent the contents of the buffer.
    */
   public String[] view()   
   {
      int lsize = size + 1 + buffer.size() + 1 + 
                  (headers == null ? 0 : headers.size() + 1); 
      String[] s = new String[lsize];
      int i = 0;

      // state flags
      s[0] = String.format("state[%d]:", new Object[] {size});

      Iterator<Integer> it = state.keySet().iterator();
      Integer id = null;
      while (it.hasNext())
      {
         StringBuffer sb = new StringBuffer();
         boolean f = false;
         id = it.next();
         Byte bw = state.get(id);
         byte b = (bw != null) ? bw.byteValue() : UNINITIALIZED;

         if (b >= ENTERED)
         {
            f = true;
            sb.append("ENTERED ");
            b -= ENTERED;
         }

         if (b >= CHANGED)
         {
            f = true;
            sb.append("CHANGED ");
            b -= CHANGED;
         }
         
         if (b >= UNINITIALIZED)
         {
            f = true;
            sb.append("UNINITIALIZED ");
            b -= UNINITIALIZED;
         }

         if (!f)         
         {
            sb.append("UNCHANGED ");
         }

         s[i + 1] = String.format(" %d: %s", new Object[] {id.intValue(), sb.toString()});
         i++;
      }

      // widgets
      s[i++] = String.format("widgets[%d]:", new Object[] {buffer.size()});
      it = buffer.keySet().iterator();
      id = null;
      Object  v  = null;
      while (it.hasNext())
      {
         id = it.next();
         v = buffer.get(id);
         if (v instanceof BaseDataType)
         {
            v = ((BaseDataType)v).toStringMessage();
         }
         s[i] = String.format(" %d: %s", new Object[] {id, v});
         i++;
      }

      // headers
      if (headers == null)
         return s;

      s[i++] = String.format("headers[%d]:", new Object[] {headers.size()});
      it = headers.keySet().iterator();
      while (it.hasNext())
      {
         id = it.next();
         v = headers.get(id);
         s[i] = String.format(" %d: %s", new Object[] {id, v});
         i++;
      }

      return s;
   }

   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      baseId     = in.readInt();
      keyCode    = in.readInt();
      down       = in.readInt();
      line       = in.readInt();
      row        = in.readInt();
      column     = in.readInt();
      widgetId   = in.readInt();
      size       = in.readInt();
      title      = (String) in.readObject();
      frameValue = (String) in.readObject();
      
      // convert the widget and header arrays back to maps
      WidgetEntry.arrayToMap((WidgetEntry[]) in.readObject(), state);
      WidgetEntry.arrayToMap((WidgetEntry[]) in.readObject(), buffer);
      WidgetEntry.arrayToMap((WidgetEntry[]) in.readObject(), headers);
      WidgetEntry.arrayToMap((WidgetEntry[]) in.readObject(), values);
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *           
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeInt(baseId);
      out.writeInt(keyCode);
      out.writeInt(down);
      out.writeInt(line);
      out.writeInt(row);
      out.writeInt(column);
      out.writeInt(widgetId);
      out.writeInt(size);
      out.writeObject(title);
      out.writeObject(frameValue);
      out.writeObject(WidgetEntry.mapToArray(state));
      out.writeObject(WidgetEntry.mapToArray(buffer));
      out.writeObject(WidgetEntry.mapToArray(headers));
      out.writeObject(WidgetEntry.mapToArray(values));
   }

   /**
    * Checks the ID to be used inside this screen buffer.
    *
    * @param    id
    *           The ID of the widget to verify.
    * @return   <code>true</code> if the widget ID can be used inside this screen buffer
    *           <code>false</code> otherwise.
    */
   private boolean isIdValid(int id)
   {
      return state.containsKey(id) || state.size() < size;
   }

   /**
    * Makes bitwise AND operation with the given widget ID state according to mask provided.  The
    * state byte after operation is storing back to the state map.
    *
    * @param    id
    *           The ID of the widget to reset status.
    * @param    mask
    *           The mask to perform bitwise and operation with the current one.
    */
   private void bitwiseAndState(int id, byte mask)
   {
      // get current status
      Byte status = state.get(id);
      // if status exists
      if (status != null)
      {
         // get it
         byte stateNew = status.byteValue();
         // change it
         stateNew &= mask;
         // and store back
         state.put(id, stateNew);
      }
   }

   /**
    * Gets the current state for the given widget ID.  Assuming the widget ID is valid for this
    * screen buffer.  Internal version of the method. 
    * 
    * @param    id
    *           The ID of the widget to which the state byte to be returned.
    *
    * @return   State of this widget or <code>UNINITIALIZED</code> if such ID does not exist in
    *           current state map.
    */
   private byte getStateInt(int id)
   {
      // get the current status
      Byte stateByte = state.get(id);
      
      return (stateByte != null) ? stateByte.byteValue() : UNINITIALIZED; 
   }

   /**
    * Stores the new state for the widget with the given widget ID.  Internal version.  If the
    * new state is UNINITIALIZED, the previous value of the widget and state, if any, is deleted.
    * 
    * @param    id
    *           The ID of the widget to which the state byte is to be stored.
    * @param    stateNew
    *           new state
    */
   private void setStateInt(int id, byte stateNew)
   {
      // state value should be kept in buffer even if widget become uninitialized or removed
      state.put(id, stateNew);
   }

   /**
    * Copies the state info map from the current state to target screen buffer.  No checking, just
    * copy existed map entries
    * 
    * @param    target
    *           a <code>ScreenBuffer</code> to copy state map info to.
    */
   private void copyState(ScreenBuffer target)
   {
      Iterator<Integer> iter = state.keySet().iterator();
      while (iter.hasNext())
      {
         Integer nextId = iter.next();
         target.state.put(nextId, this.state.get(nextId));
      }
   }
}
