/*
** Module   : ParsedDateFormat.java
** Abstract : parsed form of the date format string
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description---------------------------------
** 001 GES 20150115 Shared support for date format parsing and parse results access.
*/

package com.goldencode.p2j.util;

/**
 * This centralizes 4GL date format parsing and provides public access to the parse results as
 * simple member fields.
 */
public class ParsedDateFormat
{
   /** <code>true</code> if this format has a separator (or if it has 2 separators). */
   public boolean hasSep = true;
   
   /** Separator character between the first two date components. */
   public char sep1 = '\0';
   
   /** Separator character between the last two date components. */
   public char sep2 = '\0';
   
   /** Split format string MDY components (but not necessarily in MDY order). */
   public String[] pieces = null;
   
   /** Size of each date component. */
   public int[] size = null;
   
   /** Format string with separators converted to the / character. */
   public String normalized = null;
   
   /** Original format string. */
   public String original = null;
   
   /** <code>true</code> if parsing was successful. */
   public boolean valid = false;
   
   /**
    * Parse the given format and use the parse results to create a new instance.
    *
    * @param    fmt      
    *           The 4GL date format string to parse.           
    */
   public ParsedDateFormat(String fmt)
   {
      this.original = fmt;
      
      normalized = fmt.replaceAll("[.-]", "/");
      pieces     = normalized.split("/"); 
      
      // special case of "999999"
      if (normalized.equals("999999"))
      {
         pieces = new String[] { "99", "99", "99" };
         hasSep = false;
      }
      
      // special case of "99999999"
      if (normalized.equals("99999999"))
      {
         pieces = new String[] { "99", "99", "9999" };
         hasSep = false;
      }
      
      // the format string must have 3 parts
      if (pieces.length != 3)
      {
         String spec = "Date format %s is incomplete";
         ErrorManager.recordOrThrowError(154, String.format(spec, fmt));
         return;
      }
      
      int idx = 0;
      
      // more validity testing: each component must be non-empty
      for (int i = 0; i < 3; i++)
      {
         // empty string
         if (pieces[i].length() == 0)
         {
            date.genBadCharInFormatError(idx, fmt);
            return;
         }
         
         // keep track of the index of the next possible empty string
         idx += (pieces[i].length()) + 1;
      }
      
      int x   = 0;
      int len = normalized.length();
      
      // not done with validity test yet: the normalized format string must
      // only have '/' or '9' chars
      while (x < len)
      {
         char c = normalized.charAt(x);
         
         if (c != '/' && c != '9')
         {
            // invalid format string
            date.genBadCharInFormatError(x, fmt);
            return;
         }
         
         x++;
      }
      
      size = new int[3];
      
      // setup field sizes and separators
      for (int j = 0; j < 3; j++)
      {
         size[j] = pieces[j].length();
      }
      
      if (hasSep)
      {
         int idx2  = size[0] + 1 + size[1];
         
         sep1 = fmt.charAt(size[0]);
         sep2 = fmt.charAt(idx2);
      }
      
      valid = true;
   }
}
